//*****************************************************************************
//
// Title        : MP3stick - MP3 player
// Authors      : Michael Wolf
// File Name    : 'MP3stick.c'
// Date         : January 1, 2006
// Version      : 1.13
// Target MCU   : Atmel AVR ATmega128/1281
// Editor Tabs  : 2
//
// NOTE: The authors in no way will be responsible for damages that you
//       coul'd be using this code.
//       Use this code at your own risk.
//
//       This code is distributed under the GNU Public License
//       which can be found at http://www.gnu.org/licenses/gpl.txt
//
// Change Log
//
// Version  When        Who           What
// -------  ----        ---           ----
// 1.00     01/01/2006  Michael Wolf  Initial Release
// 1.01     12/08/2006  Michael Wolf  + Added auto power off timer when in
//                                      stop mode, default time is 60s
// 1.02     15/08/2006  Michael Wolf  - Fixed possible bug in decoding time
//                                    + Auto power off timer disabled when
//                                      charging is activated
//          17/08/2006                + Internal pull-up for charger pin enabled
//          20/08/2006                + Added low battery timer to avoid sporatic
//                                      player turn off.
//                                      Battery must be low for at least 10s to
//                                      turn the player off.
//          21/08/2006                + Added RC oscillator calibration
// 1.10     22/10/2006  Michael Wolf  + Added direct M3U playlist support
// 1.11     05/11/2006  Michael Wolf  + Added support for ID3 v1 & v2 tag
// 1.12     26/11/2006  Michael Wolf  + Added some card present checks to prevent
//                                      unnecessary card access
// 1.13			03/06/2007	Michael Wolf	* Various changes to be compatible with latest
//																			WinAVR & avr-libc version
// 1.14			04/08/2007	Michael Wolf	- Fixed bug in interrupt0 function
//																		- Fixed bug in function menu_disp where latest
//																			GCC compiler seems to produce wrong code.	
//
#define FW_VER  "v1.14"
//
//*****************************************************************************
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>

#include "MP3stick.h"
#include "hardware_conf.h"
#include "firmware_conf.h"
#include "glcd.h"
#include "vs1011.h"
#include "serial.h"
#include "fat.h"
#include "mmc.h"
#include "playlist.h"
#include "lcd_layout.h"
#include "skin.h"
#include "id3.h"

// define streams for LCD and USART
#ifdef ENABLE_SERIAL
static FILE usart_stream =
FDEV_SETUP_STREAM (usart_putchar, NULL, _FDEV_SETUP_WRITE);
#endif
static FILE lcd_stream =
FDEV_SETUP_STREAM (lcd_putchar, NULL, _FDEV_SETUP_WRITE);

volatile uint8_t actual_state = STATE_IDLE;	// storage for state machine

volatile uint8_t keytime = 0;	// key press timer
volatile uint8_t scrolltimer = 0;	// scroll speed timer
volatile uint8_t backlighttime = 0;	// backlight turn off timer
volatile uint8_t menuexittime = 50;	// menu exit timer, default 5s
volatile uint16_t auto_off_time = AUTO_OFF_TIME; // auto power off timer
volatile uint8_t batt_low_timer = 100; // timer to prevent turn off on short voltage drops
uint8_t c_backlighttime;	// hold backlight time constant in RAM
uint8_t c_backlightbright;	// hold backlight brightness constant in RAM
uint8_t c_contrast;		// hold LCD contrast in RAM

char *audiodata_pntr;	// points to audio data in buffer

volatile uint16_t actual_playlist = 0;	// actual playlist to choose titles
volatile uint16_t actual_title = 1;	// actual title to play

uint16_t playlist_entries = 0;	// number of titles in actual playlist
uint32_t playlist_length = 0;	// total length in seconds of playlist
char playlist_name[60];	// actual playlist name
uint16_t number_of_playlists = 0;	// number of playlists in root directory

uint16_t title_time = 0;	// actual title length in seconds
char titlename[60];		// actual title name
char artistname[60];		// actual artist name

volatile uint8_t is_playing = 0;		// to indicate playing state
uint8_t use_random = 0;		// use random values for title and playlist
			 // 0 = random disabled
			 // 1 = random title
			 // 2 = random playlist
			 // 3 = random title and playlist
			 // 4 = play next playlist after last title of actual list

volatile uint16_t batt_voltage = 0;	// battery voltage ADC conversion value

uint16_t last_decode_time = 0;	// to check if time display update is necessary

uint8_t artist_scroll_len;	// length of scroll line
uint8_t artist_scroll_pos;	// position of first character to scroll within string
uint8_t title_scroll_len;	// length of scroll line
uint8_t title_scroll_pos;	// position of first character to scroll within string
uint8_t playlist_scroll_len;	// length of scroll line
uint8_t playlist_scroll_pos;	// position of first character to scroll within string

// Store menu strings in program memory
uint8_t fn0[] __attribute__ ((progmem)) = T_STR0;
uint8_t fn1[] __attribute__ ((progmem)) = T_STR1;
uint8_t fn2[] __attribute__ ((progmem)) = T_STR3;
uint8_t fn3[] __attribute__ ((progmem)) = T_STR4;
uint8_t fn4[] __attribute__ ((progmem)) = T_STR5;
uint8_t fn5[] __attribute__ ((progmem)) = T_STR6;
uint8_t fn6[] __attribute__ ((progmem)) = T_STR7;
uint8_t fn7[] __attribute__ ((progmem)) = T_STR8;
uint8_t fn8[] __attribute__ ((progmem)) = T_STR9;
uint8_t fn9[] __attribute__ ((progmem)) = T_STR9;
uint8_t fn10[] __attribute__ ((progmem)) = T_STR10;
uint8_t fn11[] __attribute__ ((progmem)) = T_STR11;
uint8_t fn12[] __attribute__ ((progmem)) = T_STR11;
uint8_t fn13[] __attribute__ ((progmem)) = T_STR12;
uint8_t fn14[] __attribute__ ((progmem)) = T_STR13;
uint8_t fn15[] __attribute__ ((progmem)) = T_STR14;

// pointer array to menu strings in program memory
uint8_t *function_name[] __attribute__ ((progmem)) = {
(uint8_t *) & fn0, (uint8_t *) & fn1, (uint8_t *) & fn2,
	(uint8_t *) & fn3, (uint8_t *) & fn4, (uint8_t *) & fn5,
	(uint8_t *) & fn6, (uint8_t *) & fn7, (uint8_t *) & fn8,
	(uint8_t *) & fn9, (uint8_t *) & fn10, (uint8_t *) & fn11,
	(uint8_t *) & fn12, (uint8_t *) & fn13, (uint8_t *) & fn14,
  (uint8_t *) & fn15 };


// preset EEPROM variables
#define EEPROM __attribute__ ((section (".eeprom")))
EEPROM uint8_t ee_volume = 0;	// default full volume
EEPROM int8_t ee_balance = 100;	// default balance center
EEPROM uint8_t ee_bassboost = 0;	// default no bass boost
EEPROM uint8_t ee_randommode = 0;	// default random mode
EEPROM uint8_t ee_backlighttime = 50;	// default backlight delay time, 5s
EEPROM uint8_t ee_backlightbright = 200;	// default backlight brightness
EEPROM uint8_t ee_contrast = LCD_CONTRAST;	// default LCD contrast

//*****************************************************************************
// Function: main
//
// Description: 
//*****************************************************************************
int
main (void)
{
    OSCCAL = RC_CALIB_VALUE; // write RC oscillator calibration value

    ACSR |= _BV (ACD);		// disable Analog Comperator

    // configure SPI pins - set PB4(SS) HIGH for SPI operations!
    SPI_DIR |= _BV (MOSI) | _BV (SCK);	// config SPI pins
    SPI_PORT |= _BV (MOSI) | _BV (SCK);

    SPI_DIR |= _BV (SS);	// set SS output
    SPI_PORT |= _BV (SS);	// set SS high to enable SPI operations

    // enable SPI in master mode and set clock speed
    //  SPI clock must be at least 3MHz with 12.288MHz crystal
    SPCR = _BV (MSTR) | _BV (SPE);	// Master mode, SPI enable, clock speed MCU_XTAL/4
#if MCU_XTAL > 10000000UL
    SPCR |= _BV (SPR0);		// set clock speed MCU_XTAL/16
#endif
    SPSR |= _BV (SPI2X);	// set clock speed MCU_XTAL/2

    OCR1A = TICK_100MS;		// define TIMER1 overflow value in CTC mode
    TIMSK = _BV (OCIE1A);	// enable Timer1 output compare A interrupt 
    TCCR1B = _BV (WGM12) | _BV (CS12);	// start Timer1

    BATT_PORT |= _BV (BATT_CHG);	// enable charger pin pull-up
    batt_voltage = get_adcw (BATT_CHAN);	// get battery voltage

    // load RAM constants for backlight time and brightness
    c_backlighttime = eeprom_read_byte (&ee_backlighttime);
    c_backlightbright = eeprom_read_byte (&ee_backlightbright);
    c_contrast = eeprom_read_byte (&ee_contrast);
    
    // set default LCD contrast if EEPROM is not loaded
    if (c_contrast > 90) c_contrast = LCD_CONTRAST;

    use_random = eeprom_read_byte (&ee_randommode);
    
#ifdef ENABLE_SERIAL
    usart_init ();		// init USART
#endif
    vs1011_init ();
    vs1011_hardreset ();
    lcd_init ();		// init LCD
    lcd_contrast (c_contrast);
    set_backlight (c_backlightbright);
    text_color = BLACK;		// set default colors
    background_color = WHITE;
    lcd_cls ();			// clear LCD screen
   
    if (init_fat () == S_OK) {
        // do only if file system is present
        skin_init (0);
        skin_read_colors ();
    
        number_of_playlists = dir_list ("m3u", &playlist_file);
        get_playlist_info (actual_playlist, &playlist_entries, &playlist_length,
            playlist_name);
    }
    asm volatile ("sei");	// enable global interrupts
    go_sleep ();

    // run state machine forever
    for (;;) {

        uint8_t new_state = get_state ();	// check for key press event

        if (new_state != STATE_IDLE)
            actual_state = new_state;

        switch (actual_state) {
            default:
            case STATE_IDLE:
                break;
      
            case STATE_NEXT:
            case STATE_DOWN:
                if (!is_playing)
                    show_stopscreen ();
                next_title (1);	// get next title number
                actual_state = STATE_IDLE;
                break;
      
            case STATE_PREV:
            case STATE_UP:
                if (!is_playing)
                    show_stopscreen ();
                next_title (0);	// get previous title number
                actual_state = STATE_IDLE;
                break;
      
            case STATE_PLAY:
                if (is_playing) {
                    vs1011_hardreset ();	// reset VS1011
                    is_playing = 0;
                    show_stopscreen ();
                }
                else {
                    auto_off_time = AUTO_OFF_TIME; // reset auto power off timer
                    // play the actual title number from actual playlist
                    if (play_title () != 1) {
                        init_fat ();
                        break;
                    }
                }
                actual_state = STATE_IDLE;
                break;
      
            case STATE_STOP:
                is_playing = 1;
                actual_state = STATE_PLAY;
                break;
      
            case STATE_NEXT_PL:
                actual_playlist = (actual_playlist + 1) % number_of_playlists;
                get_playlist_info (actual_playlist, &playlist_entries,
                       &playlist_length, playlist_name);
                actual_title = 0;
                if (!is_playing)
                    show_stopscreen ();
                next_title (1);
                actual_state = STATE_IDLE;
                break;
      
            case STATE_PREV_PL:
                actual_playlist--;
                if (actual_playlist > number_of_playlists)
                    actual_playlist = number_of_playlists - 1;
                get_playlist_info (actual_playlist, &playlist_entries,
                       &playlist_length, playlist_name);
                actual_title = 0;
                if (!is_playing)
                    show_stopscreen ();
                next_title (0);
                actual_state = STATE_IDLE;
                break;
      
            case STATE_MENU:
                auto_off_time = AUTO_OFF_TIME; // reset auto power off timer
                
                // browse menu
                switch (browse_menu ()) {
            
                    case MENU_PLAY:
                        actual_state = STATE_PLAY;
                        break;
            
                    case MENU_NEXT:
                        actual_state = STATE_NEXT;
                        break;
            
                    case MENU_PREV:
                        actual_state = STATE_PREV;
                        break;
            
                    case MENU_NEXT_PL:
                        actual_state = STATE_NEXT_PL;
                        break;
            
                    case MENU_PREV_PL:
                        actual_state = STATE_PREV_PL;
                        break;
            
                    case MENU_SLEEP:
                        actual_state = STATE_GOSLEEP;
                        break;
            
                    case MENU_BOOTLOAD:
                        asm volatile("jmp 0xF000"); // start bootloader
                        break;
            
                    default:
                        actual_state = STATE_IDLE;
                        if (!is_playing)
                            show_stopscreen ();
                }
      
                if (is_playing)
                    show_titleinfo ();	// show title infos on LCD
                break;
      
            case STATE_GOSLEEP:
                go_sleep ();
                actual_state = STATE_IDLE;
                break;

        }			// end switch( actual_state)

        // turn off player if card is not present or has been removed
        // or when auto power off timer has been expired
        if ( (mmc_card_present () != 0) ||
             (auto_off_time == 0) ) {
             
            actual_state = STATE_GOSLEEP;
        }

        idle_task ();

        // turn player off if battery is low
        if ( batt_voltage < ADCW_PLAYER_OFF ) {
            
            if(batt_low_timer == 0){
                vs1011_hardreset ();	// reset VS1011
                is_playing = 0;	// stop playing 
                vs1011_sinewave_beep (4);	// send warning beeps
                actual_state = STATE_GOSLEEP;	// turn player off
            }
        }
        else
            batt_low_timer = 100; // set low battery count to 10s
            
    }				// end for(;;)
}


//*****************************************************************************
// Function: go_sleep
// Parameters: none.
// Returns: none.
//
// Description: Set the controller in power down mode and wake up on INT0
//              Measure the time key 5 is pressed.
//*****************************************************************************
void
go_sleep (void)
{
    ADCSRA = 0;			// disable ADC
    ACSR |= _BV (ACD);		// disable Analog Comperator
    lcd_sleep_in ();		// send LCD to sleep

    set_backlight (0);		// disable backlight PWM
    BL_DIR &= ~_BV (BL_SHDN);	// backlight pin input and HI-Z
    BL_PORT &= ~_BV (BL_SHDN);

    MMC_DIR |= _BV (MMC_CS);	// deselect MMC
    MMC_PORT |= _BV (MMC_CS);

    CARD_PORT &= ~_BV (CARD_DET); // disable card detection
    CARD_DIR &= ~_BV (CARD_DET);

    BATT_DIR &= ~_BV (BATT_CHG);	// set charger pins to HI-Z
    BATT_PORT &= ~_BV (BATT_CHG);	// disable charger pin pull-ups

    VS_DIR |= _BV (VS_BSYNC) | _BV (VS_CS) | _BV (VS_RESET);
    VS_PORT &= ~_BV (VS_RESET);	// keep VS1011 in reset state
    VS_PORT |= _BV (VS_CS);	// deselect VS1011

    KEY_PORT = _BV (KEY5);	// enable pull-up for power-on key
    KEY_DIR = _BV (COMM_KEYS);

    PORTE &= ~_BV (PE1);	// disable pin USART pull-ups
    PORTE &= ~_BV (PE0);

    EIMSK |= _BV (INT0);	// enable INT0
    EICRA |= _BV (ISC01);	// with falling edge detection

    asm volatile ("sei");	// enable interrupts    

pwr_down:
    TCCR1B = TIMER1_STOP;	// stop Timer1
    set_sleep_mode (SLEEP_MODE_PWR_DOWN);
    sleep_mode ();		// enter power down mode
    keytime = 0;		// reset key press time counter
    OCR1A = TICK_100MS;		// define TIMER1 overflow value in CTC mode
    TIMSK = _BV (OCIE1A);	// enable Timer1 output compare A interrupt 
    TCCR1B = _BV (WGM12) | _BV (CS12);	// start Timer1

    // wait for minimum power key press time, go back to sleep if lower
    do {
        if (bit_is_set (KEY_INP, KEY5))
            goto pwr_down;	// check if key 5 is still pressed

    } while (keytime < (uint8_t) (POWER_KEY_TIME * 10));

    asm volatile ("cli");	// disable global interrupts
    OSCCAL = RC_CALIB_VALUE; // write RC oscillator calibration value
#ifdef ENABLE_SERIAL
    usart_init ();		// init USART
#endif
    vs1011_init ();
    vs1011_hardreset ();
    lcd_init ();		// init LCD
    lcd_contrast (c_contrast);
    c_backlighttime = eeprom_read_byte (&ee_backlighttime);
    c_backlightbright = eeprom_read_byte (&ee_backlightbright);
    set_backlight (c_backlightbright);
    auto_off_time = AUTO_OFF_TIME; // reset auto power off timer
    background_color = WHITE;	// set default colors
    text_color = BLACK;
    lcd_cls ();			// clear LCD screen
    vs1011_setvolume (eeprom_read_byte (&ee_volume),
		      (int8_t) (eeprom_read_byte ((uint8_t*)&ee_balance)));
    vs1011_setbassboost (eeprom_read_byte (&ee_bassboost));
    if (init_fat () == S_OK) {
        // do only if file sysem is present
        skin_init (0);
        skin_read_colors ();
        number_of_playlists = dir_list ("m3u", &playlist_file);
        get_playlist_info (actual_playlist, &playlist_entries, &playlist_length,
               playlist_name);
    }
    else {
        number_of_playlists = 0;
    }
    is_playing = 0;
    show_stopscreen ();    
#ifdef ENABLE_SERIAL_OUTPUT
    fprintf_P (&usart_stream, PSTR ("\nMP3stick Ready\n"));
    fprintf_P (&usart_stream, PSTR (FW_VER));
    usart_putchar ('\n', NULL);
#endif
    //BATT_PORT |= _BV (BATT_CHG);	// enable charger pin pull-up
		keytime = 0;		// reset key press time counter
    asm volatile ("sei");	// enable global interrupts
}

//*****************************************************************************
// Function: set_backlight
// Parameters: PWM duty cycle
// Returns: none.
//
// Description: Turn backlight on/off with specific PWM duty cycle value
//*****************************************************************************
void
set_backlight (uint8_t duty)
{

    BL_DIR |= _BV (BL_SHDN) | _BV (BL_PWM);	// Backlight shutdown pin output
    BL_PORT |= _BV (BL_SHDN) | _BV (BL_PWM); // Turn backlight on
    if (duty > 0) {
lcd_sleep_out();
        OCR0 = duty;		// set duty cycle, higher values will increase duty cycle
        TCCR0 |= _BV (WGM01) | _BV (WGM00) | _BV (COM00)| _BV (COM01) | _BV (CS02);	// Timer0 in FastPWM
        // start backlight delay timer
        backlighttime = c_backlighttime;
    }
    else {
lcd_sleep_in();
        TCCR0 = 0;		// turn off PWM
        BL_DIR &= ~_BV (BL_SHDN);	// ensure backlight is off
        BL_PORT &= ~_BV (BL_SHDN); // Turn backlight off
    }
}


//*****************************************************************************
// Function: External INT0 service routine
// Parameters: none.
// Returns: none.
//
// Description: INT0 is used to wake up from power down mode
//*****************************************************************************
__attribute__ ((naked))
ISR (INT0_vect)
{
    asm volatile ("nop");
		asm volatile ("reti");
}

//*****************************************************************************
// Function: Timer1 Output Compare A Interupt service routine
// Parameters: none.
// Returns: none.
//
// Description: Triggered as system tick every 100ms if Timer1 is running
//*****************************************************************************
ISR (TIMER1_COMPA_vect)
{
    keytime++;
    scrolltimer++;
    batt_low_timer--;
    
    if (backlighttime) {
        backlighttime--;
        if (backlighttime == 0)
            set_backlight (0);
    }

    if (menuexittime)
        menuexittime--;
        
    if( !is_playing && auto_off_time )
        auto_off_time--;
        
}

//*****************************************************************************
// Function: get_state
// Parameters: none.
// Returns: new state from key press
//
// Description: Scan for pressed keys
//*****************************************************************************
uint8_t
get_state (void)
{

    uint8_t event = STATE_IDLE;
    static uint8_t key_press = 0;

    KEY_DIR &= ~_BV (COMM_KEYS);
    KEY_PORT &= ~_BV (COMM_KEYS);
    KEY_DIR |=
      (_BV (KEY1) | _BV (KEY2) | _BV (KEY3) | _BV (KEY4) | _BV (KEY5));
    KEY_PORT |=
      (_BV (KEY1) | _BV (KEY2) | _BV (KEY3) | _BV (KEY4) | _BV (KEY5));
    asm volatile ("nop");
    asm volatile ("nop");

#ifdef ENABLE_SERIAL
    uint8_t c;
    static uint8_t usart_chars[] = USART_CHARS;	// Tokens table for USART control

    c = usart_getchar ();

    if (c != 0)			// Check USART event
    {
        for (uint8_t i = 0; i < sizeof (usart_chars); i++) {
            if (c == usart_chars[i])	// check tokens tabele
            {
                event = i + 1;	// found
                return event;
            }
        }
    }
#endif

    // check if a key was pressed earlier
    if (key_press) {
        set_backlight (c_backlightbright);
        // check if key has been realesed
        if (bit_is_clear (KEY_INP, COMM_KEYS)) {
            // no key is down, return last key seen
            if ((keytime > 0) && (keytime < 8)) {	// max 0.8 sec for shortpress
                switch (key_press) {
                    case 0x18:
                        event = STATE_IDLE;	// full up
                        break;
                    case 0x10:
                        event = STATE_UP;	// half up
                        break;
                    case 0x01:
                        event = STATE_ENTER;	// center
                        break;
                    case 0x02:
                        event = STATE_DOWN;	// half down
                        break;
                    case 0x06:
                        event = STATE_IDLE;	// full down
                        break;
                    default:
                        event = STATE_IDLE;	// no key pressed
                }
            }
            key_press = 0;
        }
        else if ((keytime & 0x7F) >= 8)	// long keypress ? 
        {
            switch (key_press) {
                case 0x18:
                    event = STATE_IDLE;	// full up
                    break;
                case 0x10:
                    event = STATE_IDLE;	// half up
                    break;
                case 0x01:
                    event = STATE_MENU;	// center
                    break;
                case 0x02:
                    event = STATE_IDLE;	// half down
                    break;
                case 0x06:
                    event = STATE_IDLE;	// full down
                    break;
                default:
                    event = STATE_IDLE;	// no key pressed
            }
        }
    }
    else {
        // check if key is pressed
        if (bit_is_set (KEY_INP, COMM_KEYS)) {
            keytime = 0;	// reset key press timer
      
            KEY_DIR &=
              ~(_BV (COMM_KEYS) | _BV (KEY1) | _BV (KEY2) | _BV (KEY3) |
              _BV (KEY4) | _BV (KEY5));
            KEY_PORT &=
              ~(_BV (COMM_KEYS) | _BV (KEY1) | _BV (KEY2) | _BV (KEY3) |
              _BV (KEY4) | _BV (KEY5));
            KEY_DIR |= _BV (COMM_KEYS);
            KEY_PORT |=
              (_BV (KEY1) | _BV (KEY2) | _BV (KEY3) | _BV (KEY4) |
              _BV (KEY5));
      
            key_press = (~KEY_INP) & 0x1F;
        }
    }

    return event;
}


//*****************************************************************************
// Function: pump_audio
// Parameters: none.
// Returns: none.
//
// Description: Fill audio data FIFO of VS1011
//*****************************************************************************
void
pump_audio (void)
{

    int16_t read_data;

    while (bit_is_set (VS_DREQ_INP, VS_DREQ)) {

        if (audiodata_pntr == (sector_buffer + 512))	// buffer empty
        {
            audiodata_pntr = sector_buffer;	// data pointer at begin of sector buffer

            read_data = fat_read (sector_buffer, 512, &mp3_file);
      
            // if we reach the end of file play next title
            if (read_data == 0) {
                next_title (1);	// play next title
            }
        }
        vs1011_send32 (audiodata_pntr);	// send data on SPI bus
        audiodata_pntr += 32;	// advance data pointer
    }
}


//*****************************************************************************
// Function: play_title
// Parameters: none.
// Returns: Error status
//          0 = Error
//          1 = No error
//
// Description: get title infos and start audio stream
//*****************************************************************************
uint8_t
play_title (void)
{

    vs1011_softreset ();   // reset VS1011

    vs1011_setvolume (eeprom_read_byte (&ee_volume),
		      eeprom_read_byte ((uint8_t *)&ee_balance));
    vs1011_setbassboost (eeprom_read_byte (&ee_bassboost));

    is_playing = 0;		// prevent playling in case of error

    // get actual title information from playlist
    if (get_title_info (actual_playlist,
                &playlist_entries,
                &playlist_length,
                actual_title,
                &title_time,
                titlename,
                artistname,
                playlist_name) == 1
        ) {

        // read ID3 tag informations
        get_id3_tag(titlename, artistname, playlist_name, &title_time);

        // read first sector
        if (fat_read (sector_buffer, 512, &mp3_file) > 0) {
      
            show_titleinfo ();	// show title infos on LCD 
            
            audiodata_pntr = sector_buffer;	// init pointer to audio data
            pump_audio ();	// pump initial audio data to VS1011
      
            is_playing = 1;	// indicate playing state
      
            last_decode_time = 0;
      
            return 1;		// return no error
        }
    }

    return 0;			// return error

}


//*****************************************************************************
// Function: show_titleinfo
// Parameters: none.
// Returns: none.
//
// Description: show initial informations of title
//*****************************************************************************
void
show_titleinfo (void)
{

#ifdef ENABLE_SERIAL_OUTPUT
    fprintf_P (&usart_stream, PSTR ("\nNow playing:\n"));
    fprintf_P (&usart_stream, PSTR ("Playlist: %.32s\n"), playlist_name);
    fprintf_P (&usart_stream, PSTR ("Playlist Time: %0lu:%0lu:%02lu\n"),
	       (playlist_length / 60) / 60, (playlist_length / 60) % 60,
	       playlist_length % 60);
    fprintf_P (&usart_stream, PSTR ("Artist: %.60s\n"), artistname);
    fprintf_P (&usart_stream, PSTR ("Title: %.60s\n"), titlename);
    fprintf_P (&usart_stream, PSTR ("Time: %u:%02u\n"), title_time / 60,
         title_time % 60);
#endif

    // check if skin is available
    if (number_of_skins) {
        skin_image (SKIN_IMG_PLAY);
        text_color = skin_colors[SONG_NUM_FG];
        background_color = skin_colors[SONG_NUM_BG];
    }
    else
        lcd_cls ();		// clear LCD

    lcd_gotoxy (PLAYING_TITLE_XPOS, PLAYING_TITLE_YPOS);
    fprintf_P (&lcd_stream, PSTR ("%5u of %-5u"), actual_title,
	       playlist_entries);


    // check if skin is available
    if (number_of_skins) {
        text_color = skin_colors[SONG_TIME_FG];
        background_color = skin_colors[SONG_TIME_BG];
    }

    // show title time on lcd
    lcd_clrline (TITLE_TIME_YPOS);
    lcd_gotoxy (TITLE_TIME_XPOS, TITLE_TIME_YPOS);
    if (title_time !=0){
        fprintf_P (&lcd_stream, PSTR ("%u:%02u"), title_time / 60,
             title_time % 60);
    } else {
        fprintf_P (&lcd_stream, PSTR ("??:??"));
    }
    
    artist_scroll_len = strlen (artistname);	// get artist name length
    title_scroll_len = strlen (titlename);	// get title name length
    playlist_scroll_len = strlen (playlist_name);	// get playlist name length

    // limit the length
    if (artist_scroll_len > 50)
        artist_scroll_len = 50;
    if (title_scroll_len > 50)
        title_scroll_len = 50;
    if (playlist_scroll_len > 32)
        playlist_scroll_len = 32;

    artist_scroll_pos = 0;
    title_scroll_pos = 0;
    playlist_scroll_pos = 0;

    // extent name if scrolling is needed
    if (artist_scroll_len > 15) {
        strcat (artistname, " >> ");
        artist_scroll_len += 4;
    }				// if not, just print the name
    else {
        // check if skin is available
        if (number_of_skins) {
            text_color = skin_colors[ARTIST_FG];
            background_color = skin_colors[ARTIST_BG];
        }
        lcd_gotoxy (ARTIST_NAME_XPOS, ARTIST_NAME_YPOS);
        lcd_puts (artistname);
    }

    // extent name if scrolling is needed
    if (title_scroll_len > 15) {
        strcat (titlename, " >> ");
        title_scroll_len += 4;
    }				// if not, just print the name
    else {
        // check if skin is available
        if (number_of_skins) {
            text_color = skin_colors[SONG_FG];
            background_color = skin_colors[SONG_BG];
        }
        lcd_gotoxy (TITLE_NAME_XPOS, TITLE_NAME_YPOS);
        lcd_puts (titlename);
    }

    // extent name if scrolling is needed
    if (playlist_scroll_len > 15) {
        strcat (playlist_name, " >> ");
        playlist_scroll_len += 4;
    }				// if not, just print the name
    else {
        // check if skin is available
        if (number_of_skins) {
            text_color = skin_colors[PLAYLIST_FG];
            background_color = skin_colors[PLAYLIST_BG];
        }
      #ifdef CENTER_PLAYLIST
        lcd_gotoxy (LCD_POS (((15 - playlist_scroll_len) + 1) / 2),
              PLAYLIST_NAME_YPOS);
      #else
        lcd_gotoxy (PLAYLIST_NAME_XPOS, PLAYLIST_NAME_YPOS);
      #endif
        lcd_puts (playlist_name);
    }
}

//*****************************************************************************
// Function: next_title
// Parameters: 0 - previous title
//             1 - next title
// Returns: none.
//
// Description: select next title number from playlist
//*****************************************************************************
void
next_title (uint8_t dir)
{

    // check for random title and random playlist mode
    if ((use_random & 2) == 2) {
        actual_playlist = get_random () % number_of_playlists;	// get next random playlist
        get_playlist_info (actual_playlist, &playlist_entries,
               &playlist_length, playlist_name);
    }

    // check for random title from selected playlist mode
    if ((use_random & 1) == 1) {
        actual_title = get_random () % (playlist_entries + 1);	// get next random title
    }
    else {			// just increment or decrement title number
        (dir) ? actual_title++ : actual_title--;	// get new title number
    }

    // check for start of playlist, start at then end
    if (actual_title == 0) {
        // check if we need a new playlist
        if (use_random == 4) {
            actual_playlist--;
            if (actual_playlist > number_of_playlists)
                actual_playlist = number_of_playlists - 1;
            
            get_playlist_info (actual_playlist, &playlist_entries,
                   &playlist_length, playlist_name);
        }
        actual_title = playlist_entries;
    }

    // check for end of playlist, start at beginning
    if (actual_title > playlist_entries) {
        // check if we need a new playlist
        if (use_random == 4) {
            actual_playlist++;
            actual_playlist %= number_of_playlists;
            get_playlist_info (actual_playlist, &playlist_entries,
                   &playlist_length, playlist_name);
        }
        actual_title = 1;
    }

    if (is_playing)
        play_title ();		// and play if we are in playing state
}


//*****************************************************************************
// Function: get_adcw
// Parameters: none.
// Returns: none.
//
// Description: start AD conversion to read specific ADC channel
//*****************************************************************************
uint16_t
get_adcw (uint8_t chan)
{

    while ((ADCSRA & _BV (ADSC)) == _BV (ADSC));	// wait for end of running conversion

    // change conversion channel
    ADMUX |= _BV (REFS1) | _BV (REFS0);
    ADMUX &= 0xF0;		// set channel 0
    ADMUX |= chan & 0x07;	// mask selected channel

    ADCSRA |= _BV (ADEN) | _BV (ADSC);	// start 1st AD conversion
    while ((ADCSRA & _BV (ADSC)) == _BV (ADSC));	// wait for end of new conversion

    return ADCW;		// return result  
}


//*****************************************************************************
// Function: get_random
// Parameters: none.
// Returns: 16bit random value
//
// Description: generate 16bit random value
//*****************************************************************************
uint16_t
get_random (void)
{

    srandom (TCNT1 * get_adcw (3));
    return (uint16_t) random ();
}


//*****************************************************************************
// Function: idle_task
// Parameters: none.
// Returns: none.
//
// Description: run as often as possible when idle
//*****************************************************************************
void
idle_task (void)
{

    // display battery status icon
    uint8_t batt_level_color = RED;
    batt_voltage = get_adcw (BATT_CHAN);	// read battery voltage

    if (batt_voltage >= ADCW_BATT_YELLOW)
        batt_level_color = GREEN;
    else if (batt_voltage >= ADCW_BATT_RED)
        batt_level_color = YELLOW;

    lcd_gotoxy (BATT_ICON_XPOS, BATT_ICON_YPOS);

    // check if skin is available
    if (number_of_skins)
        background_color = skin_colors[BATT_BG];
    else
        background_color = WHITE;

    // check if charging is active
    if (bit_is_clear (BATT_INP, BATT_CHG)) {
        lcd_batt_icon (keytime % 16, GREEN);	// indicate charging
        
        auto_off_time = AUTO_OFF_TIME; // reset auto power off timer
    }
    else {
        // indicate battery status
        lcd_batt_icon (15 -
                 (((ADCW_BATT_GREEN -
              batt_voltage) * 15) / (ADCW_BATT_GREEN -
                   ADCW_PLAYER_OFF)),
                 batt_level_color);
    }

    if (is_playing) {

        pump_audio ();		// pump new audio data while playing

        if (scrolltimer >= SCROLL_SPEED) {
      
            scrolltimer = 0;
      
            if (artist_scroll_len > 15) {
      
                // check if skin is available
                if (number_of_skins) {
                    text_color = skin_colors[ARTIST_FG];
                    background_color = skin_colors[ARTIST_BG];
                }
            
                lcd_gotoxy (ARTIST_NAME_XPOS, ARTIST_NAME_YPOS);
                for (uint8_t i = 0; i < 15; i++) {
                    lcd_putchar (artistname
                     [(artist_scroll_pos +
                       i) % artist_scroll_len], NULL);
                }
            
                artist_scroll_pos++;
            }
      
            if (title_scroll_len > 15) {
      
                // check if skin is available
                if (number_of_skins) {
                    text_color = skin_colors[SONG_FG];
                    background_color = skin_colors[SONG_BG];
                }
            
                lcd_gotoxy (TITLE_NAME_XPOS, TITLE_NAME_YPOS);
                for (uint8_t i = 0; i < 15; i++) {
                    lcd_putchar (titlename
                     [(title_scroll_pos + i) % title_scroll_len],
                     NULL);
                }
            
                title_scroll_pos++;
            }
      
            if (playlist_scroll_len > 15) {
      
                // check if skin is available
                if (number_of_skins) {
                    text_color = skin_colors[PLAYLIST_FG];
                    background_color = skin_colors[PLAYLIST_BG];
                }
            
                lcd_gotoxy (PLAYLIST_NAME_XPOS, PLAYLIST_NAME_YPOS);
                for (uint8_t i = 0; i < 15; i++) {
                    lcd_putchar (playlist_name
                     [(playlist_scroll_pos +
                       i) % playlist_scroll_len], NULL);
                }
            
                playlist_scroll_pos++;
            }
        }
      
        uint16_t decode_time = vs1011_read (VS1011_DECODE_TIME);

        if (last_decode_time != decode_time) {

            last_decode_time = decode_time;
        
            // check if skin is available
            if (number_of_skins) {
                text_color = skin_colors[BAR_FG];
                background_color = skin_colors[BAR_BG];
            }
      
            lcd_gotoxy (PROGRESSBAR_XPOS, PROGRESSBAR_YPOS);
            lcd_bargraph (PROGRESSBAR_LEN,
              (uint8_t) ((decode_time * PROGRESSBAR_LEN) /
                   title_time), text_color,
              background_color);
      
            // check if skin is available
            if (number_of_skins) {
                text_color = skin_colors[SONG_TIME_FG];
                background_color = skin_colors[SONG_TIME_BG];
            }
      
            lcd_gotoxy (TIME_PLAYED_XPOS, TIME_PLAYED_YPOS);
            if (title_time != 0) {
                fprintf_P (&lcd_stream, PSTR ("%3u:%02u"),
                     (title_time - decode_time) / 60,
                     (title_time - decode_time) % 60);
            } else {
                fprintf_P (&lcd_stream, PSTR ("%3u:%02u"),
                     (decode_time) / 60,
                     (decode_time) % 60);
            }
        }

    }				// end if is_playing
}



//*****************************************************************************
// Function: browse_menu
// Parameters: None
// Returns: Selected menu entry
//
// Description: Browse main menu
//*****************************************************************************
uint16_t
browse_menu (void)
{

    // check if skin is available
    if (number_of_skins) {
        skin_image (SKIN_IMG_MENU);
        text_color = skin_colors[MENU_FG];
        background_color = skin_colors[MENU_BG];
    }
    else {
        lcd_cls ();		// clear LCD
        lcd_gotoxy (LCD_POS (0), LCD_LINE (0));
      
        text_color = BLUE;
				
				/*
						The following asm code replaced this line:
						lcd_puts_P (PRG_RDW ((char *) & function_name[0]));	// print menu title
						
						It's required since the new WinAVR from 20070525
						produces wrong code fragments.
				*/
				asm volatile ("ldi r24,0");
				asm volatile ("ldi r25,0");
				asm volatile ("subi r24,lo8(-(function_name))");
				asm volatile ("sbci r25,hi8(-(function_name))");
				asm volatile ("movw r30,r24");
				asm volatile ("LPM  r24,Z+");
				asm volatile ("LPM  r25,Z");
				asm volatile ("call lcd_puts_P");
				
        text_color = BLACK;
    }

    return (browse_list (1, 1, NUM_MENU_ENTRIES, menu_disp));	// run menu handler
}


//*****************************************************************************
// Function: browse_list
// Parameters:  Start index
//              Minimum index
//              Maximum index
//              Function to list content
// Returns: Selected list entry
//
// Description: Browse list
//*****************************************************************************
uint16_t
browse_list (uint16_t index,
	     uint16_t min,
	     uint16_t max,
	     void (*disp_routine) (uint16_t))
{

    uint8_t i, j = 0;
    uint8_t actual_key;
    uint8_t modeflag = 0;
    uint8_t temp = 0;
    
    // add menu items, which could be changed, here
    enum {
        VOL_CHANGE = 1,
        BAL_CHANGE,
        AMP_CHANGE,
        FREQ_CHANGE,
        BLIGHT_T_CHANGE,
        BLIGHT_B_CHANGE,
        RANDOM_CHANGE,
        CONTRAST_CHANGE,
    };

#if (MENU_EXIT_TIME > 10)
    #undef MENU_EXIT_TIME
    #define MENU_EXIT_TIME 5
    #warning "Menu exit time set to 5s!"
#endif

    menuexittime = MENU_EXIT_TIME*10;		// set auto exit time

    while (1) {
      
        // print menu list on LCD
        for (i = 0; i < 6; i++) {
      
            if (is_playing)
                pump_audio ();	// pump audio data if necessary
      
            lcd_clrline (2 + i);
            if (index + i < max) {
                // print menu entry pointer
                lcd_putchar ((i == j) ? 30 : ' ', NULL);	// offset 30 is the first character in font set
                // invert menu entry if selected for change
                if ((i == j) && (modeflag != 0)) {
                    lcd_invert = 1;
                }
                // print menu entry string
                disp_routine (index + i);
                // disable invertion
                lcd_invert = 0;
                // next entry line
                lcd_newline ();
            }
      
        }			// end for loop  

        do {
            if (is_playing)
                pump_audio ();	// pump audio data if necessary
      
            actual_key = get_state ();	// check for key press event
      
            switch (actual_key) {
            
                // change selected menu entry
                case STATE_UP:
                
                    switch(modeflag) {
                        // change volume
                        case VOL_CHANGE:
                            temp = eeprom_read_byte (&ee_volume);
                            if (temp > 0)
                                temp -= 10;
                            eeprom_write_byte (&ee_volume, temp);
                            vs1011_setvolume (temp,
                                  eeprom_read_byte ((uint8_t *)&ee_balance));
                            break;
                            
                        // change balance
                        case BAL_CHANGE:
                            temp = eeprom_read_byte ((uint8_t *)&ee_balance);
                            if (temp < 200)
                                temp += 10;
                            eeprom_write_byte ((uint8_t *)&ee_balance, temp);
                            vs1011_setvolume (eeprom_read_byte (&ee_volume),
                                  temp);
                            break;
                        
                        // change bass boost amplifier
                        case AMP_CHANGE:
                            temp = eeprom_read_byte (&ee_bassboost);
                            if (temp < 0xF0)
                                temp += 0x10;
                            eeprom_write_byte (&ee_bassboost, temp);
                            vs1011_setbassboost (temp);
                            break;
                            
                        // change bass boost frequency
                        case FREQ_CHANGE:
                            temp = eeprom_read_byte (&ee_bassboost);
                            if ((temp & 0x0F) < 0x0F)
                                temp += 0x01;
                            eeprom_write_byte (&ee_bassboost, temp);
                            vs1011_setbassboost (temp);
                            break;
                            
                        // change random mode
                        case RANDOM_CHANGE:
                            temp = eeprom_read_byte (&ee_randommode);
                            if (temp < 4)
                                temp += 1;
                            eeprom_write_byte (&ee_randommode, temp);
                            use_random = temp;
                            break;
                            
                        // change backlight auto off time
                        case BLIGHT_T_CHANGE:
                            temp = eeprom_read_byte (&ee_backlighttime);
                            if (temp < 250)
                                temp += 10;
                            eeprom_write_byte (&ee_backlighttime, temp);
                            c_backlighttime = temp;
                            break;
                            
                        // change backlight brightness
                        case BLIGHT_B_CHANGE:
                            temp = eeprom_read_byte (&ee_backlightbright);
                            if (temp < 250)
                                temp += 10;
                            eeprom_write_byte (&ee_backlightbright, temp);
                            c_backlightbright = temp;
                            break;
                            
                        // change LCD contrast
                        case CONTRAST_CHANGE:
                            temp = eeprom_read_byte (&ee_contrast);
                            if (temp < 90)
                                temp += 5;
                            eeprom_write_byte (&ee_contrast, temp);
                            c_contrast = temp;
                            lcd_contrast (temp);
                            break;
                           
                        // no entry selected, choose next entry
                        default:
                            if (index + j > min) {
                                if (j)
                                    j--;
                                else
                                    index--;
                            }
                            else {
                                index = max - 6;
                                j = 5;
                            }
                        
                    } // end switch modeflag
                    break;
            
                // change selected menu entry
                case STATE_DOWN:

                    switch(modeflag) {
                        // change volume
                        case VOL_CHANGE:
                            temp = eeprom_read_byte (&ee_volume);
                            if (temp < 100)
                                temp += 10;
                            eeprom_write_byte (&ee_volume, temp);
                            vs1011_setvolume (temp,
                                  eeprom_read_byte ((uint8_t *)&ee_balance));
                            break;
                            
                        // change balance
                        case BAL_CHANGE:
                            temp = eeprom_read_byte ((uint8_t *)&ee_balance);
                            if (temp > 0)
                                temp -= 10;
                            eeprom_write_byte ((uint8_t *)&ee_balance, temp);
                            vs1011_setvolume (eeprom_read_byte (&ee_volume),
                                  temp);
                            break;
                        
                        // change bass boost amplifier
                        case AMP_CHANGE:
                            temp = eeprom_read_byte (&ee_bassboost);
                            if ((temp & 0xF0) > 0)
                                temp -= 0x10;
                            eeprom_write_byte (&ee_bassboost, temp);
                            vs1011_setbassboost (temp);
                            break;
                        
                        // change bass boost frequency
                        case FREQ_CHANGE:
                            temp = eeprom_read_byte (&ee_bassboost);
                            if ((temp & 0x0F) > 0)
                                temp -= 0x01;
                            eeprom_write_byte (&ee_bassboost, temp);
                            vs1011_setbassboost (temp);
                            break;
                            
                        // change random mode
                        case RANDOM_CHANGE:
                            temp = eeprom_read_byte (&ee_randommode);
                            if (temp > 0)
                                temp -= 1;
                            eeprom_write_byte (&ee_randommode, temp);
                            use_random = temp;
                            break;
                        
                        // change backlight auto off time
                        case BLIGHT_T_CHANGE:
                            temp = eeprom_read_byte (&ee_backlighttime);
                            if (temp > 0)
                                temp -= 10;
                            eeprom_write_byte (&ee_backlighttime, temp);
                            c_backlighttime = temp;
                            break;
                        
                        // change backlight brightness
                        case BLIGHT_B_CHANGE:
                            temp = eeprom_read_byte (&ee_backlightbright);
                            if (temp > 30)
                                temp -= 10;
                            eeprom_write_byte (&ee_backlightbright, temp);
                            c_backlightbright = temp;
                            break;
                            
                        // change LCD contrast
                        case CONTRAST_CHANGE:
                            temp = eeprom_read_byte (&ee_contrast);
                            if (temp > 30)
                                temp -= 5;
                            eeprom_write_byte (&ee_contrast, temp);
                            c_contrast = temp;
                            lcd_contrast (temp);
                            break;
                        
                        // no entry selected, choose next entry
                        default:
                            if (index + j + 1 < max) {
                                if (j < 5)
                                    j++;
                                else
                                    index++;
                            }
                            else {
                                index = min;
                                j = 0;
                            }
                        
                    } // end switch modeflag
                    break;
            
                    // select or return menu entry
                case STATE_ENTER:
                    
                    switch(index + j) {
                        // select volume entry
                        case MENU_VOL:
                            modeflag ^= VOL_CHANGE;
                            break;
                            
                        // select balance entry
                        case MENU_BAL:
                            modeflag ^= BAL_CHANGE;
                            break;
                            
                        // select bass boost amplifier entry
                        case MENU_BASSAMP:
                            modeflag ^= AMP_CHANGE;
                            break;
                            
                        // select bass boost frequency entry
                        case MENU_BASSFREQ:
                            modeflag ^= FREQ_CHANGE;
                            break;
                            
                        // select random mode entry
                        case MENU_RAND:
                            modeflag ^= RANDOM_CHANGE;
                            break;
                            
                        // select backlight auto off time entry
                        case MENU_BLIGHT_T:
                            modeflag ^= BLIGHT_T_CHANGE;
                            break;
                            
                        // select backlight brightness entry
                        case MENU_BLIGHT_B:
                            modeflag ^= BLIGHT_B_CHANGE;
                            break;
                            
                        // select LCD contrast entry
                        case MENU_CONTRAST:
                            modeflag ^= CONTRAST_CHANGE;
                            break;
                            
                        // return non selectable entry
                        default:
                            return index + j;

                    } // end switch index+j

            }			// end switch actual key    
      
            // auto exit menu
            if (menuexittime == 0)
                return STATE_IDLE;
      
        } while (actual_key == STATE_IDLE);
      
        menuexittime = MENU_EXIT_TIME*10;	// reset auto exit time
        
        auto_off_time = AUTO_OFF_TIME; // reset auto power off timer

    }				// end while(1) loop

    return 0;			// we should never reach this return
}


//*****************************************************************************
// Function: menu_disp
// Parameters:  Entry index

// Returns: None.
//
// Description: Print menu entry strings to LCD
//*****************************************************************************
void
menu_disp (uint16_t index)
{
    if (index < NUM_MENU_ENTRIES) {

				/*
						The following asm code replaced this line:
						lcd_puts_P (PRG_RDW ((char *) & function_name[index]));	// menu print on LCD
						
						It's required since the new WinAVR from 20070525
						produces wrong code fragments.
				*/
				asm volatile ("push r24");
				asm volatile ("add r24,r28");
				asm volatile ("adc r25,r29");
				asm volatile ("subi r24,lo8(-(function_name))");
				asm volatile ("sbci r25,hi8(-(function_name))");
				asm volatile ("movw r30,r24");
				asm volatile ("LPM  r24,Z+");
				asm volatile ("LPM  r25,Z");
				asm volatile ("call lcd_puts_P");
				asm volatile ("pop r24");
				
        switch (index) {
            case MENU_VOL:
                fprintf_P (&lcd_stream, PSTR (" %u"),
                     (100 - eeprom_read_byte (&ee_volume)) / 10);
                break;
      
            case MENU_BAL:
                fprintf_P (&lcd_stream, PSTR (" %i"),
                     (int8_t) ((eeprom_read_byte ((uint8_t *)&ee_balance) -
                          100) / 10));
                break;
      
            case MENU_BASSFREQ:
                fprintf_P (&lcd_stream, PSTR (" %uHz"),
                     (eeprom_read_byte (&ee_bassboost) & 0x0F) * 10);
                break;
      
            case MENU_BASSAMP:
                fprintf_P (&lcd_stream, PSTR (" %udB"),
                     (eeprom_read_byte (&ee_bassboost) & 0xF0) >> 4);
                break;
      
            case MENU_RAND:
                switch (use_random) {
                    case 0:
                        lcd_puts_P (PSTR (" OFF"));
                        break;
            
                    case 1:
                        lcd_puts_P (PSTR (" SONG"));
                        break;
            
                    case 2:
                        lcd_puts_P (PSTR (" PL"));
                        break;
            
                    case 3:
                        lcd_puts_P (PSTR (" PL+SONG"));
                        break;
            
                    case 4:
                        lcd_puts_P (PSTR (" PL->PL"));
                        break;
            
                    default:
                        lcd_puts_P (PSTR (" OFF"));
                }
                break;
      
            case MENU_BLIGHT_T:
                fprintf_P (&lcd_stream, PSTR (" %us"),
                     eeprom_read_byte (&ee_backlighttime) / 10);
                break;
      
            case MENU_BLIGHT_B:
                fprintf_P (&lcd_stream, PSTR (" %u%%"),
                     (eeprom_read_byte (&ee_backlightbright) * 100) /
                     250);
                break;
      
        }     // end switch
    }     // end if
}


//*****************************************************************************
// Function: show_stopscreen
// Parameters: None.

// Returns: None.
//
// Description: Show player info in stopped mode
//*****************************************************************************
void
show_stopscreen (void)
{

    // check if skin is available
    if (number_of_skins) {
        skin_image (SKIN_IMG_STOP);
    }
    else {
        lcd_cls ();		// clear LCD
      
        lcd_gotoxy (LCD_POS (4), LCD_LINE (2));
        lcd_puts_P (PSTR ("MP3stick"));
      
        lcd_gotoxy (LCD_POS (5), LCD_LINE (3));
        lcd_puts_P (PSTR (FW_VER));
      
        lcd_gotoxy (LCD_POS (4), LCD_LINE (5));
        lcd_puts_P (PSTR ("Ready..."));

    }

#ifdef ENABLE_SERIAL_OUTPUT
    fprintf_P (&usart_stream, PSTR ("Stopped\n"));
#endif

}
