//*****************************************************************************
//
// Title        : MP3stick - MP3 player
// Authors      : Michael Wolf
// File Name    : 'vs1011.c'
// Date         : January 2, 2006
// Version      : 1.01
// Target MCU   : Atmel AVR ATmega128/1281
// Editor Tabs  : 2
//
// NOTE: The authors in no way will be responsible for damages that you
//       coul'd be using this code.
//       Use this code at your own risk.
//
//       This code is distributed under the GNU Public License
//       which can be found at http://www.gnu.org/licenses/gpl.txt
//
// Change Log
//
// Version  When        Who           What
// -------  ----        ---           ----
// 1.00     02/01/2006  Michael Wolf  Initial Release
// 1.01     15/08/2006  Michael Wolf  Enabled differential mode in MODE register
//
//*****************************************************************************
#include <avr/io.h>
#include <stdint.h>

#include "firmware_conf.h"
#include "hardware_conf.h"
#include "vs1011.h"

#define F_CPU   MCU_XTAL
#include <util/delay.h>

//*****************************************************************************
// Function: vs1011_sendbyte
// Parameters: Byte to send
// Returns: Byte received
//
// Description: Send one byte to VS1011 and return the received byte.
//*****************************************************************************
uint8_t
vs1011_sendbyte (uint8_t data)
{

    SPDR = data;		// send byte

    while (!(SPSR & _BV (SPIF)));	// wait until transfer finished

    return SPDR;
}

//*****************************************************************************
// Function: vs1011_init
// Parameters: none.
// Returns: none.
//
// Description: Initialize VS1011 and SPI
//*****************************************************************************
void
vs1011_init (void)
{

    VS_PORT &= ~_BV (VS_BSYNC);	// set BSYNC signal low
    VS_PORT |= _BV (VS_CS);	// deselect VS1011
    VS_PORT &= ~_BV (VS_RESET);	// keep VS1011 in reset state
    VS_DIR |= _BV (VS_BSYNC) | _BV (VS_CS) | _BV (VS_RESET);	// BSYNC, CS and RESET are outputs

    VS_DREQ_PORT |= _BV (VS_DREQ);

#ifdef VS_INIT_SPI
    // configure SPI pins - set PB4(SS) HIGH for SPI operations!
    SPI_DIR |= _BV (MOSI) | _BV (SCK);	// config SPI pins
    SPI_PORT |= _BV (MOSI) | _BV (SCK);

    SPI_DIR |= _BV (SS);	// set SS output
    SPI_PORT |= _BV (SS);	// set SS high to enable SPI operations

    // enable SPI in master mode and set clock speed
    //  SPI clock must be at least 3MHz with 12.288MHz crystal
    SPCR = _BV (MSTR) | _BV (SPE);	// Master mode, SPI enable, clock speed MCU_XTAL/4
#if MCU_XTAL > 10000000UL
    SPCR |= _BV (SPR0);		// set clock speed MCU_XTAL/16
#endif
    SPSR |= _BV (SPI2X);	// set clock speed MCU_XTAL/2
#endif
}

//*****************************************************************************
// Function: vs1011_hardreset
// Parameters: none.
// Returns: none.
//
// Description: Perform a VS1011 soft reset
//*****************************************************************************
void
vs1011_hardreset (void)
{
    asm volatile ("cli");	// disable interrupts because DREQ is an interrupt trigger

    VS_PORT &= ~_BV (VS_RESET);	// put VS1011 in reset state
    _delay_us (1);
    VS_PORT |= _BV (VS_RESET);	// put VS1011 in normal state
/*
    vs1011_nulls (2048);	// send somes NULLs

    VS_PORT &= ~_BV (VS_CS);	// select VS1011 via Chip Select

    vs1011_write (VS1011_MODE, _BV (SM_RESET));	// trigger soft reset

    _delay_us (10);		// an wait
*/
    while (!(VS_DREQ_INP & _BV (VS_DREQ)));	// wait for DREQ changes back to high

    vs1011_nulls (32);		// send somes NULLs
    vs1011_write (VS1011_CLOCKF, VS1011_CLOCKFVALUE);	// write compensation value

    VS_PORT |= _BV (VS_CS);	// deselect VS1011 via Chip Select

    asm volatile ("sei");	// disable interrupts because DREQ is an interrupt trigger
}

//*****************************************************************************
// Function: vs1011_softreset
// Parameters: none.
// Returns: none.
//
// Description: Perform a VS1011 soft reset
//*****************************************************************************
void
vs1011_softreset (void)
{
    asm volatile ("cli");	// disable interrupts because DREQ is an interrupt trigger

    vs1011_nulls (2048);	// send somes NULLs

    VS_PORT &= ~_BV (VS_CS);	// select VS1011 via Chip Select

    vs1011_write (VS1011_MODE, _BV (SM_RESET) );	// trigger soft reset

    _delay_us (10);		// an wait

    while (!(VS_DREQ_INP & _BV (VS_DREQ)));	// wait for DREQ changes back to high

    VS_PORT |= _BV (VS_CS);	// deselect VS1011 via Chip Select

    asm volatile ("sei");	// disable interrupts because DREQ is an interrupt trigger
}

//*****************************************************************************
// Function: vs1011_setvolume
// Parameters: Volume value
//             Balance value
// Returns: none.
//
// Description: Set audio volume. Create a balance effect by adding and
//              substracting the balance value from absolute volume value.
//*****************************************************************************
void
vs1011_setvolume (uint8_t volume,
		  uint8_t balance)
{

    uint16_t tempdata;

    int8_t tempbal = (int8_t) (100 - balance);

    // set volume left channel
    tempbal = -tempbal;
    tempdata = volume;
    if (tempbal > 0)
        tempdata += tempbal;

    tempdata <<= 8;
    // set volume right channel
    tempdata += volume;
    tempbal = -tempbal;
    if (tempbal > 0)
        tempdata += tempbal;

    VS_PORT &= ~_BV (VS_CS);	// select VS1011 via Chip Select

    vs1011_write (VS1011_VOL, tempdata);	// set VS1011 volume register

    vs1011_write (VS1011_MODE, _BV (SM_DIFF) );  // enable differential mode

    VS_PORT |= _BV (VS_CS);	// deselect VS1011 via Chip Select
}

//*****************************************************************************
// Function: vs1011_setbassboost
// Parameters: Amplitude value
//             Freqency limit value
// Returns: none.
//
// Description: Configure the VS1011 bass boost algorithm with specific amplitude
//              and frequency limit.
//*****************************************************************************
extern void
vs1011_setbassboost (uint8_t amp_freq)
{

    uint16_t tempdata = 0;

    tempdata += amp_freq;

    vs1011_write (VS1011_BASS, tempdata);	// set bass enhancer register
}

//*****************************************************************************
// Function: vs1011_nulls
// Parameters: Number of NULL bytes to send
// Returns: none.
//
// Description: Send N NULL bytes audio stream to VS1011
//*****************************************************************************
void
vs1011_nulls (uint16_t nNULL)
{

    for (; nNULL > 0; nNULL--) {
        vs1011_send_data (0);
    }
}

//*****************************************************************************
// Function: vs1011_write
// Parameters: Address to write
//             Data word to write
// Returns: none.
//
// Description: Write a data word to specific address
//*****************************************************************************
void
vs1011_write (uint8_t addr,
	      uint16_t data)
{

    VS_PORT &= ~_BV (VS_CS);	// select VS1011 via Chip Select

    vs1011_sendbyte (VS1011_WRITE);	// send write opcode
    vs1011_sendbyte (addr);	// send address
    vs1011_sendbyte (data >> 8);	// send MSB
    vs1011_sendbyte (data);	// send LSB

    VS_PORT |= _BV (VS_CS);	// deselect VS1011 via Chip Select
}

//*****************************************************************************
// Function: vs1011_read
// Parameters:  Address to read
// Returns: Received data word
//
// Description: Read a data word from VS1011 specific address
//*****************************************************************************
uint16_t
vs1011_read (uint8_t addr)
{

    uint16_t tempdata = 0;

    VS_PORT &= ~_BV (VS_CS);	// select VS1011 via Chip Select

    vs1011_sendbyte (VS1011_READ);	// send read opcode
    vs1011_sendbyte (addr);	// send address

    tempdata = vs1011_sendbyte (0);	// send zero dummy to receive one byte
    tempdata <<= 8;		// MSB was sent first
    tempdata += vs1011_sendbyte (0);	// send zero dummy to receive one byte

    VS_PORT |= _BV (VS_CS);	// deselect VS1011 via Chip Select

    return tempdata;		// return data word
}

//*****************************************************************************
// Function: vs1011_send_data
// Parameters:  One data byte to send
// Returns: none.
//
// Description: Send one byte of audio stream data.
//*****************************************************************************
void
vs1011_send_data (uint8_t data)
{

    asm volatile ("cli");	// disable interrupts

    VS_PORT |= _BV (VS_BSYNC);	// start BSYNC signal

    SPDR = data;		// send byte
    asm volatile ("nop");	// some delay
    asm volatile ("nop");
    asm volatile ("nop");

    VS_PORT &= ~_BV (VS_BSYNC);	// stop BSYNC signal

    while (!(SPSR & _BV (SPIF)));	// wait until transfer finished

    asm volatile ("sei");	// enable interrupts   
}

//*****************************************************************************
// Function: vs1011_send32
// Parameters: Pointer to data buffer
// Returns: none.
//
// Description: Send an 32 byte block of audio stream data
//*****************************************************************************
void
vs1011_send32 (char * pData)
{

    for (uint8_t i = 0; i < 32; i++) {
        vs1011_send_data (*pData++);	// send 32 bytes
    }
}

#ifdef SINE_BEEP
//*****************************************************************************
// Function: vs1011_sinewave_beep
// Parameters: Number of beeps to play
// Returns: none.
//
// Description: Play sine wave beeps on VS1011
//*****************************************************************************
void
vs1011_sinewave_beep (uint8_t count)
{

    static uint8_t buf[8] = { 0x53, 0xEF, 0x6E, 4,	// sine on
    0x45, 0x78, 0x69, 0x74
    };				// sine off
    vs1011_nulls (8);
    asm volatile ("cli");
    vs1011_write (VS1011_MODE, _BV (SM_TESTS));

    while (count--) {
        for (uint8_t i = 0; i < 4; i++) {
            vs1011_send_data (buf[i]);
        }
        vs1011_nulls (4);
        for (uint8_t a = 0; a < 100; a++)
            _delay_ms (1);
      
        for (uint8_t i = 4; i < 8; i++) {
            vs1011_send_data (buf[i]);
        }
        vs1011_nulls (8);
        for (uint8_t a = 0; a < 100; a++)
            _delay_ms (1);
    }
    asm volatile ("sei");
}
#endif

#ifdef VS1001_READ_MULTIPLE
//*****************************************************************************
// Function: vs1011_readmultiple
// Parameters: Address to read
//             Number of bytes to read
//             Pointer to data buffer
// Returns: none.
//
// Description: Read a specific number of bytes from VS1011
//*****************************************************************************
void
vs1011_readmultiple (uint8_t addr,
		     uint8_t count,
		     uint8_t * pData)
{

    VS_PORT &= ~_BV (VS_CS);	// select VS1011 via Chip Select

    vs1011_sendbyte (VS1011_READ);	// send read opcode
    vs1011_sendbyte (addr);	// send address

    for (; count > 0; count--) {
        *pData++ = vs1011_sendbyte (0);	// send N zero dummys and store received byte
    }

    VS_PORT |= _BV (VS_CS);	// deselect VS1011 via Chip Select
}
#endif

#ifdef VS1001_WRITE_MULTIPLE
//*****************************************************************************
// Function: vs1011_writemultiple
// Parameters: Adress to write
//             Number of bytes to write
//             Pointer to data buffer
// Returns: none.
//
// Description: Write a specific number of data bytes to VS1011
//*****************************************************************************
void
vs1011_writemultiple (uint8_t addr,
		      uint8_t count,
		      uint8_t * pData)
{

    VS_PORT &= ~_BV (VS_CS);	// select VS1011 via Chip Select

    vs1011_sendbyte (VS1011_WRITE);	// send write opcode
    vs1011_sendbyte (addr);	// send address

    for (; count > 0; count--) {
        vs1011_sendbyte (*pData++);	// send N bytes
    }

    VS_PORT |= _BV (VS_CS);	// deselect VS1011 via Chip Select
}
#endif
