//*****************************************************************************
//
// Title        : MP3stick - MP3 player
// Authors      : Michael Wolf
// File Name    : 'vs1011.h'
// Date         : January 2, 2006
// Version      : 1.00
// Target MCU   : Atmel AVR ATmega128/1281
// Editor Tabs  : 2
//
// NOTE: The authors in no way will be responsible for damages that you
//       coul'd be using this code.
//       Use this code at your own risk.
//
//       This code is distributed under the GNU Public License
//       which can be found at http://www.gnu.org/licenses/gpl.txt
//
// Change Log
//
// Version  When        Who           What
// -------  ----        ---           ----
// 1.00     02/01/2006  Michael Wolf  Initial Release
//
//*****************************************************************************
#ifndef VS1011_H
#define VS1011_H

// init SPI in vs1011_init function, comment this line to init in main()
#define VS_INIT_SPI

// VS1011 commands
#define VS1011_READ		      0x03	// Read Opcode
#define VS1011_WRITE	      0x02	// Write Opcode

// Enable multiple read and write if needed
//#define VS1001_READ_MULTIPLE
//#define VS1001_WRITE_MULTIPLE

// Enable sinewave beep function
#define SINE_BEEP

// VS1011 registers
#define VS1011_MODE         0x00	// Mode control
#define VS1011_STATUS       0x01	// Status of VS1011b
#define VS1011_BASS         0x02	// Built-in bass enhancer
#define VS1011_CLOCKF       0x03	// Clock freq + doubler
#define VS1011_DECODE_TIME  0x04	// Decode time in seconds
#define VS1011_AUDATA       0x05	// Misc. audio data
#define VS1011_WRAM         0x06	// RAM write
#define VS1011_WRAMADDR     0x07	// Base address for RAM write
#define VS1011_HDAT0        0x08	// Stream header data 0
#define VS1011_HDAT1        0x09	// Stream header data 1
#define VS1011_AIADDR       0x0A	// Start address of application
#define VS1011_VOL          0x0B	// Volume control
#define VS1011_AICTRL0      0x0C	// Application control register 0
#define VS1011_AICTRL1      0x0D	// Application control register 1
#define VS1011_AICTRL2      0x0E	// Application control register 2
#define VS1011_AICTRL3      0x0F	// Application control register 3

// Bit definition Mode Register
#define SM_DIFF         0	// Differential
#define SM_SETTOZERO1   1	// Set to zero
#define SM_RESET        2	// Soft Reset
#define SM_OUTOFWAV     3	// Jump out of WAV decoding
#define SM_SETTOZERO2   4	// Set to zero
#define SM_TESTS        5	// Allow SDI tests
#define SM_STREAM       6	// Stream mode
#define SM_SETTOZERO3   7	// Set to zero
#define SM_DACT         8	// DCLK active edge
#define SM_SDIORD       9	// SDI bit order
#define SM_SDISHARE     10	// Share SPI chip select
#define SM_SDINEW       11	// VS1002 native SPI modes
#define SM_SETTOZERO4   12	// Set to zero
#define SM_SETTOZERO5   13	// Set to zero

// Bit definition Status Register
#define SS_APDOWN1      2	// Analog internal power down
#define SS_APDOWN2      3	// Analog driver power down

// Mask definition Bass Enhancer Register
#define SB_AMPLITUDE    0xF0	// Enhancement in -1dB steps (0-15)
#define SB_FREQLIMIT    0x0F	// Low frequency limit in 10Hz steps (2-15)

// Mask and bit definitions for Stream Header data Register
#define HDAT0_EMPHAS     0x0003	// Emphasis
#define HDAT0_ORIGIN     2	// Original bit
#define HDAT0_COPY       3	// Copyright bit
#define HDAT0_EXT        0x0030	// Extension
#define HDAT0_MODE       0x00C0	// Mode
#define HDAT0_PRIV       8	// Private bit
#define HDAT0_PAD        9	// Pad bit
#define HDAT0_SAMRATE    0x0C00	// Sample rate
#define HDAT0_BITRATE    0xF000	// Bit rate
#define HDAT1_PROT       0	// Protection bit
#define HDAT1_LAYER      0x0006	// Layer
#define HDAT1_ID         0x0018	// ID
#define HDAT1_SYNC       0xFFE0	// Sync word

// calculation of master clock compensation value
#ifndef VS1011_XTAL
    #error Crystal frequency for VS1011 not defined
#endif

#if !defined(VS1011_USE_CLKDOUBLER) && (VS1011_XTAL < 15000000)
    #define VS1011_USE_CLKDOUBLER
#endif

#ifdef VS1011_USE_CLKDOUBLER
    #define TEMPCLOCKFVALUE  ((VS1011_XTAL/2000UL)+0x8000)
#else
    #define TEMPCLOCKFVALUE  (VS1011_XTAL/2000UL)
    #if TEMPCLOCKFVALUE > 32767
        #error Invalid VS1011_CLOCKFVALUE - Value greater 32767.
    #endif
#endif

#define VS1011_CLOCKFVALUE  (unsigned int)(TEMPCLOCKFVALUE)

// function prototypes
uint8_t vs1011_sendbyte (uint8_t data);
extern void vs1011_init (void);
extern void vs1011_hardreset (void);
extern void vs1011_softreset (void);
extern void vs1011_setvolume (uint8_t volume,
			      uint8_t balance);
extern void vs1011_setbassboost (uint8_t amp_freq);
extern void vs1011_nulls (uint16_t nNULL);
extern void vs1011_write (uint8_t addr,
			  uint16_t data);
extern uint16_t vs1011_read (uint8_t addr);
extern void vs1011_send_data (uint8_t data);
extern void vs1011_send32 (char * pData);

#ifdef SINE_BEEP
void vs1011_sinewave_beep (uint8_t count);
#endif

#ifdef VS1001_READ_MULTIPLE
extern void vs1011_readmultiple (uint8_t addr,
				 uint8_t count,
				 uint8_t * pData);
#endif

#ifdef VS1001_WRITE_MULTIPLE
extern void vs1011_writemultiple (uint8_t addr,
				  uint8_t count,
				  uint8_t * pData);
#endif

#endif // VS1011_H
